/*!***************************************************************************
*!
*! FILE NAME  : ReactionThread.cpp
*!
*! DESCRIPTION: This is the source code for the module `Reaction' for the
*!              Blanket Screen Saver. The code is PD.
*!
*! CLASSES    : ReactionThread
*!              MiniReactView
*!              SpeedSliderFilter
*!
*! AUTHORS    : Finn Hakansson (finn@df.lth.se)
*!              The exported interface is designed by Hirosi Lockheimer.
*!
*!***************************************************************************/

#include "Reaction.h"

#undef min
#define min(x, y) ((x) < (y) ? (x) : (y))

ReactionThread* lt = 0;
ReactionThread* mini_lt = 0;

void
module_initialize(void* theInSettings, long theInSettingsSize)
{
  if (theInSettings && theInSettingsSize)
  {
    ReactionThread::SetSpeed(*(uint32*)theInSettings);
  }
}

void
module_cleanup(void** theOutSettings, long* theOutSettingsSize)
{
  uint32* aSpeed = new uint32;
  *aSpeed = ReactionThread::Speed();
  *theOutSettings = (void*)aSpeed;
  *theOutSettingsSize = sizeof(uint32);
}

void
module_start_saving(BView* theView)
{
  lt = new ReactionThread(theView);
}

void
module_stop_saving()
{
  delete lt;
  lt = 0;
}

void
module_start_config(BView* theView)
{
  if (theView->Window()->Lock())
  {
    BFont aFont;
    font_height aFontHeight;
    aFont.SetSize(14);
    aFont.GetHeight(&aFontHeight);
    int aHeight = 0;

    BRect aFrame = theView->Bounds();
    aFrame.bottom = aFontHeight.ascent + aFontHeight.descent +
                    aFontHeight.leading;
    BStringView* aName = new BStringView(aFrame, B_EMPTY_STRING, "Reaction");
    aName->SetFont(&aFont);
    aName->SetAlignment(B_ALIGN_CENTER);
    theView->AddChild(aName);
    aName->SetViewColor(theView->ViewColor());

    aFont.SetSize(9.0);
    aFont.GetHeight(&aFontHeight);
    BRect aFrame2 = theView->Bounds();
    aFrame2.top = aFrame.bottom + 2.0;
    aFrame2.bottom = aFrame2.top + aFontHeight.ascent +
                     aFontHeight.descent + aFontHeight.leading;

    BStringView* aVersionString =
      new BStringView(aFrame2, B_EMPTY_STRING, "Version 1.0 (Aug 10, 1998).");
    aVersionString->SetFont(&aFont);
    aVersionString->SetAlignment(B_ALIGN_CENTER);
    theView->AddChild(aVersionString);
    aVersionString->SetViewColor(theView->ViewColor());

    aFont.SetSize(10.0);
    aFont.GetHeight(&aFontHeight);
    BRect aFrame3 = theView->Bounds();
    aFrame3.top = aFrame2.bottom + 2.0;
    aFrame3.bottom = aFrame3.top + aFontHeight.ascent +
                     aFontHeight.descent + aFontHeight.leading;
    BStringView* aNameString =
      new BStringView(aFrame3, B_EMPTY_STRING,
                      "by Finn Hakansson (finn@df.lth.se)");
    aNameString->SetFont(&aFont);
    aNameString->SetAlignment(B_ALIGN_CENTER);
    theView->AddChild(aNameString);
    aNameString->SetViewColor(theView->ViewColor());

    BRect aFrame4 = theView->Bounds();
    aFrame4.top = aFrame3.bottom + 5.0;
    aFrame4.bottom = aFrame4.top;
    BSlider* aSpeedSlider = new BSlider(aFrame4, "aSpeedSlider",
                                        "Reaction Speed:",
                                        new BMessage('fha1'), 1, 100);
    aSpeedSlider->SetValue(ReactionThread::Speed());
    rgb_color aColor;
    aColor.red = 30;
    aColor.green = 30;
    aColor.blue = 255;
    aColor.alpha = 0;
    aSpeedSlider->SetBarColor(aColor);
    rgb_color* aFillColor = new rgb_color;
    aFillColor->red = 255;
    aFillColor->green = 0;
    aFillColor->blue = 0;
    aFillColor->alpha = 0;
    aSpeedSlider->UseFillColor(true, (const rgb_color*)aFillColor);
    aSpeedSlider->SetTarget(theView);
    aSpeedSlider->SetHashMarkCount(8);
    aSpeedSlider->SetHashMarks(B_HASH_MARKS_NONE);
    aSpeedSlider->SetLimitLabels("Slow", "Fast");
    aSpeedSlider->ResizeToPreferred();
    theView->AddChild(aSpeedSlider);
    aHeight = aSpeedSlider->Bounds().Height();
    theView->AddFilter(new SpeedSliderFilter(aSpeedSlider));

    BRect aPreViewFrame = theView->Bounds();
    aPreViewFrame.top = aFrame4.bottom + aHeight + 10;
    aPreViewFrame.left += 10.0;
    aPreViewFrame.right -= 10.0;
    aPreViewFrame.bottom -= 10.0;
    MiniReactView *aPreView = new MiniReactView(aPreViewFrame);
    theView->AddChild(aPreView);
    aPreView->SetViewColor(0, 0, 0, 0);
    aPreView->SetLowColor(0, 0, 0, 0);
    aPreView->Invalidate();

    theView->Window()->UpdateIfNeeded();
    theView->Window()->Unlock();
    mini_lt = new ReactionThread(aPreView);
  }
}

void
module_stop_config()
{
  delete mini_lt;
  mini_lt = 0;
}

/****************** ReactionThread DEFINITION SECTION ***********************/

uint32 ReactionThread::mySpeed = 20;

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::ReactionThread()
*#
*# DESCRIPTION  : Constructor.
*#
*#--------------------------------------------------------------------------*/
ReactionThread::ReactionThread(BView* theView)
  : mySemaphore(0),
    myThread(0),
    myView(theView),
    myScreenWidth(0),
    myScreenHeight(0),
    myXMiddle(0),
    myYMiddle(0),
    myRealRadius(2),
    myMaxRadius(0),
    myXPosition(0),
    myYPosition(0),
    myColorIndex(NUMBER_OF_COLORS - 1),
    iAmFresh(true),
    iAmInitialized(false)
{
  srand((long)system_time());
  myView->Window()->Show();

  myColors[0].red = 50;
  myColors[0].green = 255;
  myColors[0].blue = 0;
  myColors[0].alpha = 0;

  myColors[1].red = 130;
  myColors[1].green = 0;
  myColors[1].blue = 255;
  myColors[1].alpha = 0;

  myColors[2].red = 0;
  myColors[2].green = 255;
  myColors[2].blue = 255;
  myColors[2].alpha = 0;

  myColors[3].red = 70;
  myColors[3].green = 255;
  myColors[3].blue = 70;
  myColors[3].alpha = 0;

  myColors[4].red = 255;
  myColors[4].green = 255;
  myColors[4].blue = 0;
  myColors[4].alpha = 0;

  myColors[5].red = 240;
  myColors[5].green = 0;
  myColors[5].blue = 255;
  myColors[5].alpha = 0;

  if (myView->Window()->Lock())
  {
    BRect aBoundsRect = myView->Bounds();
    myScreenWidth = aBoundsRect.IntegerWidth();
    myScreenHeight = aBoundsRect.IntegerHeight();

    myReaction = new uint8*[myScreenHeight];
    if (!myReaction)
    {
      myView->Window()->Unlock();
      return;
    }
    for (int i = 0; i < myScreenHeight; i++)
    {
      myReaction[i] = new uint8[myScreenWidth];
      if (!myReaction[i])
      {
        myView->Window()->Unlock();
        return;
      }
    }
    myXMiddle = myScreenWidth / 2;
    myYMiddle = myScreenHeight / 2;
    myMaxRadius = min(myXMiddle - 1, myYMiddle - 1);
    mySemaphore = create_sem(1, "ReactSem");
    myThread = spawn_thread(proc, "ReactThread", B_LOW_PRIORITY, this);
    resume_thread(myThread);
    myView->Window()->Unlock();
  }
}

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::~ReactionThread()
*#
*# DESCRIPTION  : Destructor.
*#
*#--------------------------------------------------------------------------*/
ReactionThread::~ReactionThread()
{
  acquire_sem(mySemaphore);
  delete_sem(mySemaphore);
  mySemaphore = 0;
  long value = 0;
  wait_for_thread(myThread, &value);
  myThread = 0;
}

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::proc()
*#
*# DESCRIPTION  : The spawned thread's main entrance point.
*#
*#--------------------------------------------------------------------------*/
long
ReactionThread::proc(void* theData)
{
  ReactionThread* lt = (ReactionThread*)theData;
  sem_id aSemaphore = lt->mySemaphore;

  while (acquire_sem(aSemaphore) == B_NO_ERROR)
  {
    lt->Save();
    release_sem(aSemaphore);

    if (acquire_sem(aSemaphore) != B_NO_ERROR)
    {
      return B_NO_ERROR;
    }
    else
    {
      snooze(20000.0);
      release_sem(aSemaphore);
    }
  }

  return B_NO_ERROR;
}

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::Save()
*#
*# DESCRIPTION  : Do some saving.
*#
*#--------------------------------------------------------------------------*/
void
ReactionThread::Save()
{
  if (myView->Window()->Lock())
  {
    if (iAmFresh)
    {
      InitScreen();
    }
    for (int i = 0; i < mySpeed; i++)
    {
      if (React())
      {
        break;
      }
    }
    myView->Sync();
    myView->Window()->Unlock();
  }
}

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::React()
*#
*# DESCRIPTION  : 
*#
*#--------------------------------------------------------------------------*/
bool
ReactionThread::React()
{
  double r = rand();
  myXPosition = cos(r) * myRealRadius + myXMiddle;
  myYPosition = sin(r) * myRealRadius + myYMiddle;

  while (!Reacted())
  {
    myXPosition += (int)((uint32)rand() % 3) - 1;
    if (OutsideRealRadius())
    {
      if (myXPosition < myXMiddle)
      {
        myXPosition++;
      }
      else
      {
        myXPosition--;
      }
    }
    myYPosition += (int)((uint32)rand() % 3) - 1;
    if (OutsideRealRadius())
    {
      if (myYPosition < myYMiddle)
      {
        myYPosition++;
      }
      else
      {
        myYPosition--;
      }
    }
  }

  if (OnRealRadius())
  {
    myRealRadius += 1.0;
    if (myRealRadius >= myMaxRadius)
    {
      iAmFresh = true;
      return true;
    }
  }

  // Draw a dot in the view.
  myView->SetHighColor(myColors[myColorIndex]);
  BPoint p(myXPosition, myYPosition);
  myView->StrokeLine(p, p, B_SOLID_HIGH);
  myReaction[myYPosition][myXPosition] = 1;
  return false;
}

/*----------------------------------------------------------------------------
*#
*# FUNCTION NAME: ReactionThread::Clear()
*#
*# DESCRIPTION  : Clear the whole reaction.
*#
*#--------------------------------------------------------------------------*/
void
ReactionThread::Clear()
{
  for (int i = 0; i < myScreenHeight; i++)
  {
    for (int j = 0; j < myScreenWidth; j++)
    {
      myReaction[i][j] = 0;
    }
  }
}

bool
ReactionThread::Reacted()
{
  return ((myReaction[myYPosition - 1][myXPosition - 1]) ||
          (myReaction[myYPosition - 1][myXPosition])     ||
          (myReaction[myYPosition - 1][myXPosition + 1]) ||
          (myReaction[myYPosition][myXPosition - 1])     ||
          (myReaction[myYPosition][myXPosition + 1])     ||
          (myReaction[myYPosition + 1][myXPosition - 1]) ||
          (myReaction[myYPosition + 1][myXPosition])     ||
          (myReaction[myYPosition + 1][myXPosition + 1]));
}

void
ReactionThread::InitScreen()
{
  Clear();
  myReaction[myYMiddle][myXMiddle] = 1;
  myView->SetHighColor(0, 0, 0, 0);
  myView->FillRect(myView->Bounds(), B_SOLID_HIGH);
  myColorIndex = (myColorIndex + 1) % NUMBER_OF_COLORS;
  myView->SetHighColor(myColors[myColorIndex]);
  BPoint p(myXMiddle, myYMiddle);
  myView->StrokeLine(p, p);
  iAmFresh = false;
  iAmInitialized = true;
  myRealRadius = 3;
}

bool
ReactionThread::OutsideRealRadius()
{
  return ((int)distance(myXMiddle, myYMiddle, myXPosition, myYPosition) > myRealRadius);
}

bool
ReactionThread::OnRealRadius()
{
  return ((int)distance(myXMiddle, myYMiddle, myXPosition, myYPosition) >= myRealRadius);
}

inline double
ReactionThread::distance(double x1, double y1, double x2, double y2)
{
  return sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
}

void
ReactionThread::Draw(BRect theUpdateRect)
{
  myView->Window()->Lock();
  myView->SetHighColor(0, 0, 0, 0);
  myView->FillRect(theUpdateRect, B_SOLID_HIGH);
  myView->SetHighColor(myColors[myColorIndex]);

  for (int y = theUpdateRect.top; y < theUpdateRect.bottom; y++)
  {
    for (int x = theUpdateRect.left; x < theUpdateRect.right; x++)
    {
      if (myReaction[y][x])
      {
        myView->StrokeLine(BPoint(x, y), BPoint(x, y), B_SOLID_HIGH);
      }
    }
  }

  myView->Sync();
  myView->Window()->Unlock();

}

inline bool
ReactionThread::isInitialized()
{
  return iAmInitialized;
}

uint32
ReactionThread::Speed()
{
  return mySpeed;
}

void
ReactionThread::SetSpeed(uint32 theNewSpeed)
{
  mySpeed = theNewSpeed;
}

/****************** MiniReactView DEFINITION SECTION ************************/

MiniReactView::MiniReactView(BRect theFrame)
  : BView(theFrame, B_EMPTY_STRING, B_FOLLOW_ALL, B_WILL_DRAW)
{
  // Empty.
}

void
MiniReactView::Draw(BRect theUpdateRect)
{
  if (mini_lt->isInitialized())
  {
    mini_lt->Draw(theUpdateRect);
  }
}

/****************** SpeedSliderFilter DEFINITION SECTION ********************/

SpeedSliderFilter::SpeedSliderFilter(BSlider* theSpeedSlider)
  : BMessageFilter(B_ANY_DELIVERY, B_ANY_SOURCE),
    mySpeedSliderRef(theSpeedSlider)
{
  // Empty.
}

filter_result
SpeedSliderFilter::Filter(BMessage* theMessage, BHandler** theHandler)
{
#pragma unused (theHandler)
  switch (theMessage->what)
  {
  case 'fha1':
    ReactionThread::SetSpeed((uint32)mySpeedSliderRef->Value());
    break;
  }
  return B_DISPATCH_MESSAGE;
}

/****************** END OF FILE Reaction.cpp ********************************/
